# -*- coding: utf-8 -*-
"""
tabledataextractor.tests.test_table_table.py
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Test main table object.
These tests depend on the input from csv working properly.
Juraj Mavračić (jm2111@cam.ac.uk)
"""

import unittest
import logging

from tabledataextractor import Table
from tabledataextractor.input import from_csv
from tabledataextractor.output.print import print_table
from tabledataextractor.table.algorithms import find_cc4, find_cc1_cc2
import numpy as np
import os

log = logging.getLogger(__name__)


class TableCC4(Table):
    """Derivative of Table used to isolate testing for CC4"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def labels(self):
        """Cell labels. Python List"""
        temp = np.empty_like(self._pre_cleaned_table, dtype="<U60")
        temp[:, :] = '/'
        temp[self._cc4] = 'CC4'
        return temp


class TestCC4(unittest.TestCase):

    def do_table(self, input_path, expected):
        log.debug("Test CC4, Table: {}".format(input_path))
        table = TableCC4(input_path, use_spanning_cells=False, use_header_extension=False)
        table.print()
        result = find_cc4(table)
        log.debug("Result = {}".format(result))
        self.assertTupleEqual(expected, result)

    def test_cc4_1(self):
        input_path = './tests/data/table_example1.csv'
        expected = (6,7)
        self.do_table(input_path, expected)

    def test_cc4_2(self):
        """
        Here, the true result would actually be (4,5).
        However, in _find_cc4() we choose the criterion that the data region starts, from below, when at least half of
        the row is non-empty.
        """
        input_path = './tests/data/table_example2.csv'
        expected = (3,5)
        self.do_table(input_path, expected)

    def test_cc4_3(self):
        input_path = './tests/data/table_example3.csv'
        expected = (7,10)
        self.do_table(input_path, expected)

    def test_cc4_4(self):
        input_path = './tests/data/table_example4.csv'
        expected = (2,5)
        self.do_table(input_path, expected)

    def test_cc4_5(self):
        input_path = './tests/data/table_example5.csv'
        expected = (3,5)
        self.do_table(input_path, expected)

    def test_cc4_6(self):
        input_path = './tests/data/table_example6.csv'
        expected = (11,3)
        self.do_table(input_path, expected)

    def test_cc4_7(self):
        input_path = './tests/data/table_example7.csv'
        expected = (8,10)
        self.do_table(input_path, expected)


class TableCC1CC2(Table):
    """Derivative of Table used to isolate testing for CC1 and CC2"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def labels(self):
        """Cell labels. Python List"""
        temp = np.empty_like(self._pre_cleaned_table, dtype="<U60")
        temp[:, :] = '/'
        temp[self._cc4] = 'CC4'
        temp[self._cc1] = 'CC1'
        temp[self._cc2] = 'CC2'
        return temp


class TableCC3(Table):
    """Derivative of Table used to isolate testing for CC3"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def labels(self):
        """Cell labels. Python List"""
        temp = np.empty_like(self._pre_cleaned_table, dtype="<U60")
        temp[:, :] = '/'
        temp[self._cc4] = 'CC4'
        temp[self._cc1] = 'CC1'
        temp[self._cc2] = 'CC2'
        temp[self._cc3] = 'CC3'
        return temp


class TestCC1CC2(unittest.TestCase):

    def do_table(self, input_path, expected):
        log.debug("Test CC1 & CC2, Table: {}".format(input_path))
        table = TableCC1CC2(input_path, use_spanning_cells=False, use_header_extension=False)
        table.print()
        result = find_cc1_cc2(table, find_cc4(table), table.pre_cleaned_table)
        log.debug("Result = {}".format(result))
        self.assertTupleEqual(expected, result)

    def test_cc1_cc2_1(self):
        input_path = './tests/data/table_example1.csv'
        expected = (0,1),(1,1)
        self.do_table(input_path, expected)

    def test_cc1_cc2_2(self):
        """Lack of row header section on the LHS"""
        input_path = './tests/data/table_example2.csv'
        expected = (0,0),(2,0)
        self.do_table(input_path, expected)

    def test_cc1_cc2_3(self):
        input_path = './tests/data/table_example3.csv'
        expected = (1,0),(2,0)
        self.do_table(input_path, expected)

    def test_cc1_cc2_4(self):
        input_path = './tests/data/table_example4.csv'
        expected = (0,0),(0,0)
        self.do_table(input_path, expected)

    def test_cc1_cc2_5(self):
        input_path = './tests/data/table_example5.csv'
        expected = (0,0),(1,0)
        self.do_table(input_path, expected)

    def test_cc1_cc2_6(self):
        """Table is composed of two tables"""
        input_path = './tests/data/table_example6.csv'
        expected = (1,0),(1,0)
        self.do_table(input_path, expected)

    def test_cc1_cc2_7(self):
        input_path = './tests/data/table_example7.csv'
        expected = (1,0),(2,0)
        self.do_table(input_path, expected)


class TestCC3(unittest.TestCase):

    def do_table(self, input_path, expected):
        log.debug("Test CC3, Table: {}".format(input_path))
        table = TableCC3(input_path, use_spanning_cells=False, use_header_extension=False)
        result = table._cc3
        log.debug("Result = {}".format(result))
        table.print()
        self.assertTupleEqual(expected, result)

    def test_cc3_1(self):
        input_path = './tests/data/table_example1.csv'
        expected = (2,2)
        self.do_table(input_path, expected)

    def test_cc3_2(self):
        """Lack of row header section on the LHS"""
        input_path = './tests/data/table_example2.csv'
        expected = (3,1)
        self.do_table(input_path, expected)

    def test_cc3_3(self):
        input_path = './tests/data/table_example3.csv'
        expected = (3,1)
        self.do_table(input_path, expected)

    def test_cc3_4(self):
        input_path = './tests/data/table_example4.csv'
        expected = (1,1)
        self.do_table(input_path, expected)

    def test_cc3_5(self):
        input_path = './tests/data/table_example5.csv'
        expected = (2,1)
        self.do_table(input_path, expected)

    def test_cc3_6(self):
        """Table is composed of two tables"""
        input_path = './tests/data/table_example6.csv'
        expected = (2,1)
        self.do_table(input_path, expected)

    def test_cc3_7(self):
        input_path = './tests/data/table_example7.csv'
        expected = (4,1)
        self.do_table(input_path, expected)


class TestDuplicateLabelPrefixing(unittest.TestCase):

    def do_table(self, input_path, expected_path):
        log.debug("Test duplicate label prefixing: {}".format(input_path))
        table = Table(input_path, use_footnotes=False, use_spanning_cells=False, use_header_extension=False)
        print_table(table.raw_table)
        print_table(table._pre_cleaned_table)
        result = table._pre_cleaned_table.tolist()
        expected = from_csv.read(expected_path).tolist()
        self.assertListEqual(expected, result)

    def test_table_8(self):
        """Prefixing in column header"""
        input_path = './tests/data/table_example8.csv'
        expected_path = './tests/data/table_example8b.csv'
        self.do_table(input_path, expected_path)

    def test_table_9(self):
        """Prefixing in row header"""
        input_path = './tests/data/table_example9.csv'
        expected_path = './tests/data/table_example9b.csv'
        self.do_table(input_path, expected_path)

    def test_table_10(self):
        """Prefixing will destroy the labelling"""
        input_path = './tests/data/table_example10.csv'
        expected_path = './tests/data/table_example10b.csv'
        self.do_table(input_path, expected_path)

    def test_table_11(self):
        """Prefixing is not performed, column header"""
        input_path = './tests/data/table_example11.csv'
        expected_path = './tests/data/table_example11b.csv'
        self.do_table(input_path, expected_path)

    def test_table_12(self):
        """Prefixing is not performed, row header"""
        input_path = './tests/data/table_example12.csv'
        expected_path = './tests/data/table_example12b.csv'
        self.do_table(input_path, expected_path)


class TestTableLabels(unittest.TestCase):

    def do_table(self, input_path, expected):
        log.debug("Test complete table labelling: {}".format(input_path))
        table = Table(input_path, use_footnotes=False, use_spanning_cells=False, use_header_extension=False)
        print(repr(table))
        result = table.labels.tolist()
        self.assertListEqual(expected, result)

    def test_table_1(self):
        input_path = './tests/data/table_example1.csv'
        expected = [['TableTitle', 'StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['/',    'StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['Note', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['Note', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['Note', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['Note', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['Note', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        self.do_table(input_path, expected)

    def test_table_2(self):
        """Lack of row header section on the LHS"""
        input_path = './tests/data/table_example2.csv'
        expected = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['Note', 'Note', 'Note', '/', '/', '/']]
        self.do_table(input_path, expected)

    def test_table_3(self):
        input_path = './tests/data/table_example3.csv'
        expected = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'],
                    ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        self.do_table(input_path, expected)

    def test_table_4(self):
        input_path = './tests/data/table_example4.csv'
        expected = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        self.do_table(input_path, expected)

    def test_table_5(self):
        input_path = './tests/data/table_example5.csv'
        expected = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        self.do_table(input_path, expected)

    def test_table_6(self):
        """Table is composed of two tables"""
        input_path = './tests/data/table_example6.csv'
        expected = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'],
                    ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data']]
        self.do_table(input_path, expected)

    def test_table_7(self):
        input_path = './tests/data/table_example7.csv'
        expected = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'],
                    ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader & FNref', 'ColHeader & FNref', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader & FNref', 'ColHeader & FNref'],
                    ['FNref', '/', '/', '/', '/', '/', '/', '/', '/', '/', '/'],
                    ['RowHeader & FNref', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['RowHeader & FNref', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['FNprefix & FNtext', 'Note', '/', '/', '/', '/', '/', '/', '/', '/', '/'],
                    ['FNprefix', 'FNtext', '/', '/', '/', '/', '/', '/', '/', '/', '/'],
                    ['Note', '/', '/', '/', '/', '/', '/', '/', '/', '/', '/'],
                    ['FNprefix & FNtext', '/', '/', '/', '/', '/', '/', '/', '/', '/', '/']]

        self.do_table(input_path, expected)


class TestCategorizationTable(unittest.TestCase):

    def do_table(self, input_path, expected):
        log.debug("Test Categorization Table: {}".format(input_path))
        table = Table(input_path, use_footnotes=False, use_spanning_cells=False, use_header_extension=False, standardize_empty_data=False)
        print(repr(table))
        result = table.category_table
        self.assertListEqual(expected, result)

    def test_table_1(self):
        input_path = os.path.join(os.path.dirname(__file__), 'data', 'table_example1.csv')
        expected = [['4.64', ['This study'], ['Rutile', 'a = b (A)']], ['2.99', ['This study'], ['Rutile', 'c (A)']], ['0.305', ['This study'], ['Rutile', 'u']], ['3.83', ['This study'], ['Anatase', 'a = b (A)']], ['9.62', ['This study'], ['Anatase', 'c (A)']], ['0.208', ['This study'], ['Anatase', 'u']], ['4.67', ['GGA [25]'], ['Rutile', 'a = b (A)']], ['2.97', ['GGA [25]'], ['Rutile', 'c (A)']], ['0.305', ['GGA [25]'], ['Rutile', 'u']], ['3.80', ['GGA [25]'], ['Anatase', 'a = b (A)']], ['9.67', ['GGA [25]'], ['Anatase', 'c (A)']], ['0.207', ['GGA [25]'], ['Anatase', 'u']], ['4.63', ['GGA [26]'], ['Rutile', 'a = b (A)']], ['2.98', ['GGA [26]'], ['Rutile', 'c (A)']], ['0.305', ['GGA [26]'], ['Rutile', 'u']], ['', ['GGA [26]'], ['Anatase', 'a = b (A)']], ['', ['GGA [26]'], ['Anatase', 'c (A)']], ['', ['GGA [26]'], ['Anatase', 'u']], ['', ['HF [27]'], ['Rutile', 'a = b (A)']], ['', ['HF [27]'], ['Rutile', 'c (A)']], ['', ['HF [27]'], ['Rutile', 'u']], ['3.76', ['HF [27]'], ['Anatase', 'a = b (A)']], ['9.85', ['HF [27]'], ['Anatase', 'c (A)']], ['0.202', ['HF [27]'], ['Anatase', 'u']], ['4.594', ['Expt. [23]'], ['Rutile', 'a = b (A)']], ['2.958', ['Expt. [23]'], ['Rutile', 'c (A)']], ['0.305', ['Expt. [23]'], ['Rutile', 'u']], ['3.785', ['Expt. [23]'], ['Anatase', 'a = b (A)']], ['9.514', ['Expt. [23]'], ['Anatase', 'c (A)']], ['0.207', ['Expt. [23]'], ['Anatase', 'u']]]
        self.do_table(input_path, expected)

    def test_table_2(self):
        """Table without a row header."""
        input_path = './tests/data/table_example2.csv'
        expected = [['2.990', ['4.640'], ['2006', 'Government transfers', 'Implicit transfer rates1 %']], ['0.305', ['4.640'], ['2006', 'Government transfers', 'Shares %']], ['3.830', ['4.640'], ['2007', 'Government transfers', 'Average $ constant 2007']], ['9.620', ['4.640'], ['2007', 'Government transfers', 'Implicit transfer rates1 %']], ['0.208', ['4.640'], ['2007', 'Government transfers', 'Shares %']]]
        self.do_table(input_path, expected)

    def test_table_3(self):
        input_path = './tests/data/table_example3.csv'
        expected = [['3735', ['Norway'], ['Million dollar', '2007']], ['4006', ['Norway'], ['Million dollar', '2008']], ['4081', ['Norway'], ['Million dollar', '2009']], ['4580', ['Norway'], ['Million dollar', '2010*']], ['4936', ['Norway'], ['Million dollar', '2011*']], ['0.95', ['Norway'], ['Percentage of GNI', '2007']], ['0.89', ['Norway'], ['Percentage of GNI', '2008']], ['1.06', ['Norway'], ['Percentage of GNI', '2009']], ['1.1', ['Norway'], ['Percentage of GNI', '2010*']], ['1', ['Norway'], ['Percentage of GNI', '2011*']], ['2562', ['Denmark'], ['Million dollar', '2007']], ['2803', ['Denmark'], ['Million dollar', '2008']], ['2810', ['Denmark'], ['Million dollar', '2009']], ['2871', ['Denmark'], ['Million dollar', '2010*']], ['2981', ['Denmark'], ['Million dollar', '2011*']], ['0.81', ['Denmark'], ['Percentage of GNI', '2007']], ['0.82', ['Denmark'], ['Percentage of GNI', '2008']], ['0.88', ['Denmark'], ['Percentage of GNI', '2009']], ['0.91', ['Denmark'], ['Percentage of GNI', '2010*']], ['0.86', ['Denmark'], ['Percentage of GNI', '2011*']], ['2669', ['Australia'], ['Million dollar', '2007']], ['2954', ['Australia'], ['Million dollar', '2008']], ['2762', ['Australia'], ['Million dollar', '2009']], ['3826', ['Australia'], ['Million dollar', '2010*']], ['4799', ['Australia'], ['Million dollar', '2011*']], ['0.32', ['Australia'], ['Percentage of GNI', '2007']], ['0.32', ['Australia'], ['Percentage of GNI', '2008']], ['0.29', ['Australia'], ['Percentage of GNI', '2009']], ['0.32', ['Australia'], ['Percentage of GNI', '2010*']], ['0.35', ['Australia'], ['Percentage of GNI', '2011*']], ['320', ['New Zealand'], ['Million dollar', '2007']], ['348', ['New Zealand'], ['Million dollar', '2008']], ['309', ['New Zealand'], ['Million dollar', '2009']], ['342', ['New Zealand'], ['Million dollar', '2010*']], ['429', ['New Zealand'], ['Million dollar', '2011*']], ['0.27', ['New Zealand'], ['Percentage of GNI', '2007']], ['0.3', ['New Zealand'], ['Percentage of GNI', '2008']], ['0.28', ['New Zealand'], ['Percentage of GNI', '2009']], ['0.26', ['New Zealand'], ['Percentage of GNI', '2010*']], ['0.28', ['New Zealand'], ['Percentage of GNI', '2011*']], ['104206', ['OECD/DAC1'], ['Million dollar', '2007']], ['121954', ['OECD/DAC1'], ['Million dollar', '2008']], ['119778', ['OECD/DAC1'], ['Million dollar', '2009']], ['128465', ['OECD/DAC1'], ['Million dollar', '2010*']], ['133526', ['OECD/DAC1'], ['Million dollar', '2011*']], ['0.27', ['OECD/DAC1'], ['Percentage of GNI', '2007']], ['0.3', ['OECD/DAC1'], ['Percentage of GNI', '2008']], ['0.31', ['OECD/DAC1'], ['Percentage of GNI', '2009']], ['0.32', ['OECD/DAC1'], ['Percentage of GNI', '2010*']], ['0.31', ['OECD/DAC1'], ['Percentage of GNI', '2011*']]]
        self.do_table(input_path, expected)

    def test_table_4(self):
        input_path = './tests/data/table_example4.csv'
        expected = [['1647218', ['2003'], ['Short messages/thousands']], ['24.3', ['2003'], ['A Change %']], ['347', ['2003'], [' Short messages/subscription']], ['2314', ['2003'], ['Multimedia messages/thousands']], ['', ['2003'], ['B Change %']], ['2193498', ['2004'], ['Short messages/thousands']], ['33.2', ['2004'], ['A Change %']], ['439', ['2004'], [' Short messages/subscription']], ['7386', ['2004'], ['Multimedia messages/thousands']], ['219.2', ['2004'], ['B Change %']]]
        self.do_table(input_path, expected)

    def test_table_5(self):
        input_path = './tests/data/table_example5.csv'
        expected = [['1647218', ['2003'], ['', 'Short messages/thousands']], ['24.3', ['2003'], ['Short messages/thousands', 'Change %']], ['347', ['2003'], ['', ' Short messages/subscription']], ['2314', ['2003'], ['', 'Multimedia messages/thousands']], ['', ['2003'], ['Multimedia messages/thousands', 'Change %']], ['2193498', ['2004'], ['', 'Short messages/thousands']], ['33.2', ['2004'], ['Short messages/thousands', 'Change %']], ['439', ['2004'], ['', ' Short messages/subscription']], ['7386', ['2004'], ['', 'Multimedia messages/thousands']], ['219.2', ['2004'], ['Multimedia messages/thousands', 'Change %']]]
        self.do_table(input_path, expected)

    def test_table_7(self):
        input_path = './tests/data/table_example7.csv'
        expected = [['3735', ['Norwaya)'], ['Million dollar', '2007']], ['4081', ['Norwaya)'], ['Million dollar', '2009']], ['4006', ['Norwaya)'], ['Million dollar', '2008']], ['4580', ['Norwaya)'], ['Million dollar', '2010*']], ['4936', ['Norwaya)'], ['Million dollar', '2011*']], ['0.95', ['Norwaya)'], ['Percentage of GNI', '2007']], ['0.89', ['Norwaya)'], ['Percentage of GNI', '2008']], ['1.06', ['Norwaya)'], ['Percentage of GNI', '2009']], ['1.1', ['Norwaya)'], ['Percentage of GNI', '2010*']], ['1', ['Norwaya)'], ['Percentage of GNI', '2011*']], ['2562', ['Denmark'], ['Million dollar', '2007']], ['2810', ['Denmark'], ['Million dollar', '2009']], ['2803', ['Denmark'], ['Million dollar', '2008']], ['2871', ['Denmark'], ['Million dollar', '2010*']], ['2981', ['Denmark'], ['Million dollar', '2011*']], ['0.81', ['Denmark'], ['Percentage of GNI', '2007']], ['0.82', ['Denmark'], ['Percentage of GNI', '2008']], ['0.88', ['Denmark'], ['Percentage of GNI', '2009']], ['0.91', ['Denmark'], ['Percentage of GNI', '2010*']], ['0.86', ['Denmark'], ['Percentage of GNI', '2011*']], ['2669', ['Australia'], ['Million dollar', '2007']], ['2762', ['Australia'], ['Million dollar', '2009']], ['2954', ['Australia'], ['Million dollar', '2008']], ['3826', ['Australia'], ['Million dollar', '2010*']], ['4799', ['Australia'], ['Million dollar', '2011*']], ['0.32', ['Australia'], ['Percentage of GNI', '2007']], ['0.32', ['Australia'], ['Percentage of GNI', '2008']], ['0.29', ['Australia'], ['Percentage of GNI', '2009']], ['0.32', ['Australia'], ['Percentage of GNI', '2010*']], ['0.35', ['Australia'], ['Percentage of GNI', '2011*']], ['320', ['New Zealand'], ['Million dollar', '2007']], ['309', ['New Zealand'], ['Million dollar', '2009']], ['348', ['New Zealand'], ['Million dollar', '2008']], ['342', ['New Zealand'], ['Million dollar', '2010*']], ['429', ['New Zealand'], ['Million dollar', '2011*']], ['0.27', ['New Zealand'], ['Percentage of GNI', '2007']], ['0.3', ['New Zealand'], ['Percentage of GNI', '2008']], ['0.28', ['New Zealand'], ['Percentage of GNI', '2009']], ['0.26', ['New Zealand'], ['Percentage of GNI', '2010*']], ['0.28', ['New Zealand'], ['Percentage of GNI', '2011*']], ['104206', ['OECD/DAC 1'], ['Million dollar', '2007']], ['119778', ['OECD/DAC 1'], ['Million dollar', '2009']], ['121954', ['OECD/DAC 1'], ['Million dollar', '2008']], ['128465', ['OECD/DAC 1'], ['Million dollar', '2010*']], ['133526', ['OECD/DAC 1'], ['Million dollar', '2011*']], ['0.27', ['OECD/DAC 1'], ['Percentage of GNI', '2007']], ['0.3', ['OECD/DAC 1'], ['Percentage of GNI', '2008']], ['0.31', ['OECD/DAC 1'], ['Percentage of GNI', '2009']], ['0.32', ['OECD/DAC 1'], ['Percentage of GNI', '2010*']], ['0.31', ['OECD/DAC 1'], ['Percentage of GNI', '2011*']]]
        self.do_table(input_path, expected)

    def test_table_8(self):
        input_path = './tests/data/table_example8.csv'
        expected = [['1647218', ['2003'], ['', 'Short messages/thousands']], ['24.3', ['2003'], ['Short messages/thousands', 'Change %']], ['347', ['2003'], ['', 'Short messages/subscription']], ['2314', ['2003'], ['', 'Multimedia messages/thousands']], ['', ['2003'], ['Multimedia messages/thousands', 'Change %']], ['2193498', ['2004'], ['', 'Short messages/thousands']], ['33.2', ['2004'], ['Short messages/thousands', 'Change %']], ['439', ['2004'], ['', 'Short messages/subscription']], ['7386', ['2004'], ['', 'Multimedia messages/thousands']], ['219.2', ['2004'], ['Multimedia messages/thousands', 'Change %']]]
        self.do_table(input_path, expected)

    def test_table_9(self):
        input_path = './tests/data/table_example9.csv'
        expected = [['1647218', ['', 'Short messages/thousands'], ['2003']], ['2193498', ['', 'Short messages/thousands'], ['2004']], ['24.3', ['Short messages/thousands', 'Change %'], ['2003']], ['33.2', ['Short messages/thousands', 'Change %'], ['2004']], ['347', ['', 'Short messages/subscription'], ['2003']], ['439', ['', 'Short messages/subscription'], ['2004']], ['2314', ['', 'Multimedia messages/thousands'], ['2003']], ['7386', ['', 'Multimedia messages/thousands'], ['2004']], ['', ['Multimedia messages/thousands', 'Change %'], ['2003']], ['219.2', ['Multimedia messages/thousands', 'Change %'], ['2004']]]
        self.do_table(input_path, expected)

    def test_table_10(self):
        """Table where prefixing destroys the layout"""
        input_path = './tests/data/table_example10.csv'
        expected = [['', [''], ['', 'Short messages/thousands']], ['A', [''], ['Short messages/thousands', 'Change %']], ['', [''], ['', ' Short messages/subscription']], ['', [''], ['', 'Multimedia messages/thousands']], ['B', [''], ['Multimedia messages/thousands', 'Change %']], ['1647218', ['2003'], ['', 'Short messages/thousands']], ['24.3', ['2003'], ['Short messages/thousands', 'Change %']], ['347', ['2003'], ['', ' Short messages/subscription']], ['2314', ['2003'], ['', 'Multimedia messages/thousands']], ['', ['2003'], ['Multimedia messages/thousands', 'Change %']], ['2193498', ['2004'], ['', 'Short messages/thousands']], ['33.2', ['2004'], ['Short messages/thousands', 'Change %']], ['439', ['2004'], ['', ' Short messages/subscription']], ['7386', ['2004'], ['', 'Multimedia messages/thousands']], ['219.2', ['2004'], ['Multimedia messages/thousands', 'Change %']]]
        self.do_table(input_path, expected)

    def test_table_11(self):
        input_path = './tests/data/table_example11b.csv'
        expected = [['1647218', ['2003'], ['', 'Short messages/thousands']], ['24.3', ['2003'], ['Category A', 'Change %']], ['347', ['2003'], ['', ' Short messages/subscription']], ['2314', ['2003'], ['', 'Multimedia messages/thousands']], ['', ['2003'], ['Category B', 'Change %']], ['2193498', ['2004'], ['', 'Short messages/thousands']], ['33.2', ['2004'], ['Category A', 'Change %']], ['439', ['2004'], ['', ' Short messages/subscription']], ['7386', ['2004'], ['', 'Multimedia messages/thousands']], ['219.2', ['2004'], ['Category B', 'Change %']]]
        self.do_table(input_path, expected)

    def test_table_12(self):
        input_path = './tests/data/table_example12.csv'
        expected = [['1647218', ['', 'Short messages/thousands'], ['2003']], ['2193498', ['', 'Short messages/thousands'], ['2004']], ['24.3', ['Category A', 'Change %'], ['2003']], ['33.2', ['Category A', 'Change %'], ['2004']], ['347', ['', 'Short messages/subscription'], ['2003']], ['439', ['', 'Short messages/subscription'], ['2004']], ['2314', ['', 'Multimedia messages/thousands'], ['2003']], ['7386', ['', 'Multimedia messages/thousands'], ['2004']], ['', ['Category B', 'Change %'], ['2003']], ['219.2', ['Category B', 'Change %'], ['2004']]]
        self.do_table(input_path, expected)

    def test_table_13(self):
        """Coordination numbers"""
        input_path = './tests/data/table_example13.csv'
        expected = [['1.82 ± 0.3', ['NAsAs'], ['As40Te60']], ['1.07 ± 0.3', ['NAsAs'], ['As30Cu10Te60']], ['1.13 ± 0.3', ['NAsAs'], ['As20Cu20Te60']], ['1.18 ± 0.3', ['NAsAs'], ['As15Cu25Te60']], ['0.79 ± 0.3', ['NAsAs'], ['As10Cu30Te60']], ['-', ['NAsCu'], ['As40Te60']], ['0.37 ± 0.2', ['NAsCu'], ['As30Cu10Te60']], ['0.71 ± 0.25', ['NAsCu'], ['As20Cu20Te60']], ['0.85 ± 0.25', ['NAsCu'], ['As15Cu25Te60']], ['0.51 ± 0.3', ['NAsCu'], ['As10Cu30Te60']], ['1.34 ± 0.3', ['NAsTe'], ['As40Te60']], ['1.58 ± 0.3', ['NAsTe'], ['As30Cu10Te60']], ['1.80 ± 0.3', ['NAsTe'], ['As20Cu20Te60']], ['1.94 ± 0.4', ['NAsTe'], ['As15Cu25Te60']], ['2.76 ± 0.6', ['NAsTe'], ['As10Cu30Te60']], ['-', ['NCuAs'], ['As40Te60']], ['1.11 ± 0.6', ['NCuAs'], ['As30Cu10Te60']], ['0.71 ± 0.25', ['NCuAs'], ['As20Cu20Te60']], ['0.51 ± 0.15', ['NCuAs'], ['As15Cu25Te60']], ['0.17 ± 0.1', ['NCuAs'], ['As10Cu30Te60']], ['-', ['NCuCu'], ['As40Te60']], ['1.07 ± 0.6', ['NCuCu'], ['As30Cu10Te60']], ['1.69 ± 0.5', ['NCuCu'], ['As20Cu20Te60']], ['1.83 ± 0.5', ['NCuCu'], ['As15Cu25Te60']], ['3.11 ± 0.7', ['NCuCu'], ['As10Cu30Te60']], ['-', ['NCuTe'], ['As40Te60']], ['2.32 ± 0.6', ['NCuTe'], ['As30Cu10Te60']], ['2.01 ± 0.6', ['NCuTe'], ['As20Cu20Te60']], ['2.46 ± 0.6', ['NCuTe'], ['As15Cu25Te60']], ['2.17 ± 0.5', ['NCuTe'], ['As10Cu30Te60']], ['0.89', ['NTeAs'], ['As40Te60']], ['0.79 ± 0.25', ['NTeAs'], ['As30Cu10Te60']], ['0.60 ± 0.2', ['NTeAs'], ['As20Cu20Te60']], ['0.49 ± 0.2', ['NTeAs'], ['As15Cu25Te60']], ['0.46 ± 0.1', ['NTeAs'], ['As10Cu30Te60']], ['-', ['NTeCu'], ['As40Te60']], ['0.39 ± 0.2', ['NTeCu'], ['As30Cu10Te60']], ['0.67 ± 0.2', ['NTeCu'], ['As20Cu20Te60']], ['1.03 ± 0.25', ['NTeCu'], ['As15Cu25Te60']], ['1.08 ± 0.25', ['NTeCu'], ['As10Cu30Te60']], ['1.01', ['NTeTe'], ['As40Te60']], ['1.14 ± 0.2', ['NTeTe'], ['As30Cu10Te60']], ['1.43 ± 0.2', ['NTeTe'], ['As20Cu20Te60']], ['1.44 ± 0.2', ['NTeTe'], ['As15Cu25Te60']], ['1.65 ± 0.2', ['NTeTe'], ['As10Cu30Te60']], ['3.16 ± 0.3', ['NAsAs + NAsTe'], ['As40Te60']], ['2.65 ± 0.3', ['NAsAs + NAsTe'], ['As30Cu10Te60']], ['2.93 ± 0.3', ['NAsAs + NAsTe'], ['As20Cu20Te60']], ['3.12 ± 0.3', ['NAsAs + NAsTe'], ['As15Cu25Te60']], ['3.55 ± 0.6', ['NAsAs + NAsTe'], ['As10Cu30Te60']], ['1.90 ± 0.2', ['NTeAs + NTeTe'], ['As40Te60']], ['1.93 ± 0.2', ['NTeAs + NTeTe'], ['As30Cu10Te60']], ['2.03 ± 0.2', ['NTeAs + NTeTe'], ['As20Cu20Te60']], ['1.93 ± 0.2', ['NTeAs + NTeTe'], ['As15Cu25Te60']], ['2.11 ± 0.2', ['NTeAs + NTeTe'], ['As10Cu30Te60']], ['3.16 ± 0.3', ['NAs'], ['As40Te60']], ['3.02 ± 0.3', ['NAs'], ['As30Cu10Te60']], ['3.64 ± 0.3', ['NAs'], ['As20Cu20Te60']], ['3.97 ± 0.3', ['NAs'], ['As15Cu25Te60']], ['4.06 ± 0.6', ['NAs'], ['As10Cu30Te60']], ['-', ['NCu'], ['As40Te60']], ['4.50 ± 1.0', ['NCu'], ['As30Cu10Te60']], ['4.41 + 0.5', ['NCu'], ['As20Cu20Te60']], ['4.80 ± 0.5', ['NCu'], ['As15Cu25Te60']], ['5.45 ± 0.6', ['NCu'], ['As10Cu30Te60']], ['1.90 ± 0.2', ['NTe'], ['As40Te60']], ['2.32 ± 0.2', ['NTe'], ['As30Cu10Te60']], ['2.70 ± 0.2', ['NTe'], ['As20Cu20Te60']], ['2.96 ± 0.2', ['NTe'], ['As15Cu25Te60']], ['3.19 ± 0.3', ['NTe'], ['As10Cu30Te60']]]
        self.do_table(input_path, expected)


class TestSingleRowColumnTable(unittest.TestCase):

    def do_table(self, input_path, expected, expected_t, expected_cat, expected_cat_t):
        log.debug("Test single column/row table: {}".format(input_path))

        # Table labels
        table = Table(input_path, use_footnotes=False, use_spanning_cells=False, use_header_extension=False)
        print(repr(table))
        result = table.labels.tolist()
        self.assertListEqual(expected, result)

        # Category table
        result = table.category_table
        self.assertListEqual(expected_cat, result)

        # Inverted table labels
        table.transpose()
        print(repr(table))
        result = table.labels.tolist()
        self.assertListEqual(expected_t, result)

        # Inverted category table
        result = table.category_table
        self.assertListEqual(expected_cat_t, result)

    def test_table_1(self):
        input_path = './tests/data/te_01.csv'
        expected = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_t = [['StubHeader', 'ColHeader'],
                      ['RowHeader', 'Data'],
                      ['RowHeader', 'Data'],
                      ['RowHeader', 'Data'],
                      ['RowHeader', 'Data'],
                      ['RowHeader', 'Data'],
                      ['RowHeader', 'Data']]
        expected_cat = [['4.64', ['This study'], ['A']], ['2.99', ['This study'], ['B']], ['0.305', ['This study'], ['C']], ['3.83', ['This study'], ['D']], ['9.62', ['This study'], ['E']], ['0.208', ['This study'], ['F']]]
        expected_cat_t = [['4.64', ['A'], ['This study']], ['2.99', ['B'], ['This study']], ['0.305', ['C'], ['This study']], ['3.83', ['D'], ['This study']], ['9.62', ['E'], ['This study']], ['0.208', ['F'], ['This study']]]
        self.do_table(input_path, expected, expected_t, expected_cat, expected_cat_t)

    def test_table_2(self):
        input_path = './tests/data/te_02.csv'
        expected = [['StubHeader', 'ColHeader'],
                    ['RowHeader', 'Data'],
                    ['RowHeader', 'Data'],
                    ['RowHeader', 'Data'],
                    ['RowHeader', 'Data'],
                    ['RowHeader', 'Data'],
                    ['RowHeader', 'Data']]
        expected_t = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                      ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_cat = [['4.64', ['A'], ['This study']], ['2.99', ['B'], ['This study']], ['0.305', ['C'], ['This study']], ['3.83', ['D'], ['This study']], ['9.62', ['E'], ['This study']], ['0.208', ['F'], ['This study']]]
        expected_cat_t = [['4.64', ['This study'], ['A']], ['2.99', ['This study'], ['B']], ['0.305', ['This study'], ['C']], ['3.83', ['This study'], ['D']], ['9.62', ['This study'], ['E']], ['0.208', ['This study'], ['F']]]
        self.do_table(input_path, expected, expected_t, expected_cat, expected_cat_t)

    def test_table_3(self):
        input_path = './tests/data/te_03.csv'
        expected = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'],
                    ['/', 'StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                    ['/', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data'],
                    ['Note', 'RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data']]
        expected_t = [['TableTitle', 'StubHeader', 'ColHeader', 'ColHeader'],
                      ['Note', 'RowHeader', 'Data', 'Data'],
                      ['Note', 'RowHeader', 'Data', 'Data'],
                      ['Note', 'RowHeader', 'Data', 'Data'],
                      ['Note', 'RowHeader', 'Data', 'Data'],
                      ['Note', 'RowHeader', 'Data', 'Data'],
                      ['Note', 'RowHeader', 'Data', 'Data']]
        expected_cat = [['B', ['A'], ['Test3']], ['C', ['A'], ['Test4']], ['D', ['A'], ['Test5']], ['E', ['A'], ['Test6']], ['F', ['A'], ['Test7']], ['2.99', ['4.64'], ['Test3']], ['0.305', ['4.64'], ['Test4']], ['3.83', ['4.64'], ['Test5']], ['9.62', ['4.64'], ['Test6']], ['0.208', ['4.64'], ['Test7']]]
        expected_cat_t = [['A', ['Test2'], ['']], ['4.64', ['Test2'], ['This study']], ['B', ['Test3'], ['']], ['2.99', ['Test3'], ['This study']], ['C', ['Test4'], ['']], ['0.305', ['Test4'], ['This study']], ['D', ['Test5'], ['']], ['3.83', ['Test5'], ['This study']], ['E', ['Test6'], ['']], ['9.62', ['Test6'], ['This study']], ['F', ['Test7'], ['']], ['0.208', ['Test7'], ['This study']]]
        self.do_table(input_path, expected, expected_t, expected_cat, expected_cat_t)


if __name__ == '__main__':
    unittest.main()





# -*- coding: utf-8 -*-
"""
Tools for parsing the table based on regular expressions.
"""

import logging
import re
import numpy as np

log = logging.getLogger(__name__)


class CellParser:
    """
    :param pattern: Regular expression pattern which defines the cell parser. Use `grouping`, since matching strings will be returned explicitly.
    :type pattern: str

    """
    def __init__(self, pattern):

        log.debug('Initialization of CellParser with regex pattern: "{}"'.format(pattern))
        assert isinstance(pattern, str)
        self.pattern = pattern

    def parse(self, table, method='match'):
        """
        Inputs a table and yields a tuple with the index of the next matching cell, as well as the string that
        was matched.

        :param method:  `search`, `match` or `fullmatch`; Python `Regular expressions <https://docs.python.org/3.6/library/re.html>`_
        :type method: str
        :param table: Input table to be parsed
        :type table: numpy.array
        :yield: (int, int, str) with index of cells and the strings of the groups that were matched

        """

        # check if table is of correct type
        assert isinstance(table, np.ndarray)

        result = None
        prog = re.compile(self.pattern)

        # check the dimensionality of the array
        if table.ndim == 2:
            for row_index, row in enumerate(table):
                for column_index, cell in enumerate(row):
                    if method == 'match':
                        result = prog.match(cell)
                    elif method == 'fullmatch':
                        result = prog.fullmatch(cell)
                    elif method == 'search':
                        result = prog.search(cell)
                    if result:
                        yield row_index, column_index, result.groups()
        elif table.ndim == 1:
            for row_index, row in enumerate(table):
                if method == 'match':
                    result = prog.match(row)
                elif method == 'fullmatch':
                    result = prog.fullmatch(row)
                elif method == 'search':
                    result = prog.search(row)
                if result:
                    yield row_index, result.groups()

    def cut(self, table, method='match'):
        """
        Inputs a table and yields a tuple with the index of the next matching cell, as well as a string
        that is obtained from the original string by cutting out the match string.

        :param method:  `search`, `match` or `fullmatch`; see Python `Regular expressions <https://docs.python.org/3.6/library/re.html>`_
        :type method: str
        :param table: Input table to be parsed, of type 'numpy.ndarray'
        :type table: numpy.array
        :yield: (int, int, str) with index of cells and the strings of the groups that were matched

        """
        # check if table is of correct type
        assert isinstance(table, np.ndarray)

        prog = re.compile(self.pattern)
        for result in self.parse(table, method):
            yield result[0], result[1], prog.sub("", table[result[:2]])

    def replace(self, table, repl, method='match'):
        """
        Inputs a table and yields a tuple with the index of the next matching cell, as well as a string
        that is obtained from the original string by cutting out the match string and replacing it with another string.

        :param method: `search`, `match` or `fullmatch`; see Python `Regular expressions <https://docs.python.org/3.6/library/re.html>`_
        :type method: str
        :param table: Input table to be parsed
        :type table: numpy.array
        :param repl: Replacement string that will be included instead of the patters
        :type repl: str
        :yield: (int, int, str) with index of cells and the strings of the groups that were matched

        """

        # check if table is of correct type
        assert isinstance(table, np.ndarray)

        prog = re.compile(self.pattern)
        for result in self.parse(table, method):
            yield result[0], result[1], prog.sub(repl, table[result[:2]])


class StringParser:
    """
    :param pattern: Regular expression pattern that defines the string parser.
    :type pattern: str

    """
    def __init__(self, pattern):
        assert isinstance(pattern, str)
        self.pattern = pattern

    def parse(self, string, method='match'):
        """
        Inputs a string and returns `True` if pattern matches.

        :param string: Input string
        :param method: `search`, `match` or `fullmatch`; see Python `Regular expressions <https://docs.python.org/3.6/library/re.html>`_
        :type string: str
        :type method: str
        :return: True/False
        """

        # check if string is of correct type
        assert isinstance(string, str)

        result = None
        prog = re.compile(self.pattern)

        if method == 'match':
            result = prog.match(string)
        elif method == 'fullmatch':
            result = prog.fullmatch(string)
        elif method == 'search':
            result = prog.search(string)
        if result:
            return True
        else:
            return False

    def cut(self, string):
        """
        Inputs a string and returns the same string with the pattern cut out

        :param string: Input string
        :type string: str
        :return: string with `pattern` cut out
        """

        # check if string is of correct type
        assert isinstance(string, str)

        prog = re.compile(self.pattern)
        result = prog.sub(string, "")
        return result



# -*- coding: utf-8 -*-
"""
tabledataextractor.tests.test_input_from_csv.py
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Test csv table output parser
Ed Beard (ejb207@cam.ac.uk)
"""


import unittest
import logging
import os

from tabledataextractor import Table
from tabledataextractor.output.to_csv import write_to_csv


log = logging.getLogger(__name__)


class TestOutputCsv(unittest.TestCase):

    def do_conversion(self, filename):
        """ Helper function to convert from input to output"""

        # Paths to csv files
        in_path = os.path.join(os.path.dirname(__file__), 'data', filename)
        out_path = os.path.join(os.path.dirname(__file__), 'data', 'temp_' + filename)

        # Load input csv as tde_table and string
        with open(in_path, 'r', encoding='utf-8') as f:
            in_string = f.read()
        table = Table(in_path)

        # Output csv and import as string
        write_to_csv(table.raw_table, out_path)
        with open(in_path, 'r', encoding='utf-8') as f:
            out_string = f.read()
        os.remove(out_path)

        self.assertEqual(in_string, out_string)

    def test_table_example1(self):
        self.do_conversion('table_example1.csv')

    def test_table_double_quotes(self):
        self.do_conversion('table_double_quotes.csv')

    def test_all_example_tables(self):
        for file in os.listdir(os.path.join(os.path.dirname(__file__), 'data')):
            self.do_conversion(file)









# -*- coding: utf-8 -*-
"""
tabledataextractor.tests.test_input_from_csv.py
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Test table parsers.
Juraj Mavračić (jm2111@cam.ac.uk)
Ed Beard (ejb207@cam.ac.uk)
"""

import unittest
import logging
import os

from tabledataextractor import Table

log = logging.getLogger(__name__)


class TestInputCsv(unittest.TestCase):

    def test_input_with_commas(self):

        path = os.path.join(os.path.dirname(__file__), 'data', 'table_commas.csv')
        table = Table(path)

        # Normal table
        table.print()
        row_0_cell_6 = table.raw_table[0][6]
        self.assertEqual(row_0_cell_6, 'PCE (η, %)')

        row_1_cell_0 = table.raw_table[1][0]
        self.assertEqual(row_1_cell_0, '2H–MoS2 (hydrothermal, 200 °C)')

        row_2_cell_0 = table.raw_table[2][0]
        self.assertEqual(row_2_cell_0, '1T–MoS2 (hydrothermal, 180 °C)')

    def test_input_with_double_quotes(self):
        path = os.path.join(os.path.dirname(__file__), 'data', 'table_double_quotes.csv')
        table = Table(path)

        row_1_cell_0 = table.raw_table[1][0]
        self.assertEqual(row_1_cell_0, 'N719 "cell')

        row_3_cell_0 = table.raw_table[3][0]
        self.assertEqual(row_3_cell_0, 'Hy"brid cell')

    def test_input_with_single_quotes(self):
        path = os.path.join(os.path.dirname(__file__), 'data', 'table_single_quotes.csv')
        table = Table(path)

        row_1_cell_0 = table.raw_table[1][0]
        self.assertEqual(row_1_cell_0, "N719 'cell")

        row_2_cell_0 = table.raw_table[2][0]
        self.assertEqual(row_2_cell_0, "N74'9 cell")

        row_2_cell_0 = table.raw_table[2][0]
        self.assertEqual(row_2_cell_0, "N74'9 cell")

        row_3_cell_2 = table.raw_table[3][2]
        self.assertEqual(row_3_cell_2, "13'.8")

    def test_input_with_blank_lines(self):
        path = os.path.join(os.path.dirname(__file__), 'data', 'table_empty_lines.csv')
        table = Table(path)

        # Normal table
        row_0_cell_6 = table.raw_table[0][6]
        self.assertEqual(row_0_cell_6, 'PCE (η, %)')

        row_1_cell_0 = table.raw_table[1][0]
        self.assertEqual(row_1_cell_0, '2H–MoS2 (hydrothermal, 200 °C)')

        row_2_cell_0 = table.raw_table[2][0]
        self.assertEqual(row_2_cell_0, '1T–MoS2 (hydrothermal, 180 °C)')

    def test_table_with_broken_rows(self):
        path = os.path.join(os.path.dirname(__file__), 'data', 'table_broken_row.csv')
        table = Table(path)

        self.assertEqual(len(table.category_table), 10)




# -*- coding: utf-8 -*-
"""
Outputs the table to a Pandas DataFrame.
"""

import pandas as pd


def to_pandas(table):
    """
    Creates a `Pandas <http://pandas.pydata.org/>`_ `DataFrame` object from a :class:`~tabledataextractor.table.table.Table` object.

    :param table: Input table
    :type table: ~tabledataextractor.table.table.Table
    :return: :class:`pandas.DataFrame`
    """
    index_row = pd.MultiIndex.from_arrays(table.row_header.T)
    index_col = pd.MultiIndex.from_arrays(table.col_header)
    df = pd.DataFrame(columns=index_col, index=index_row, data=table.data)
    return df


def find_multiindex_level(row_number, column_number, df):
    """
    Helping function for ``_build_category_table()``.
    Finds the `Pandas` `MultiIndex level` in a given `Pandas` `DataFrame`, for a particular data value.
    """
    result_index = []
    if hasattr(df.index, 'codes'):
        for i, codes in enumerate(df.index.codes):
            result_index.append(df.index.levels[i][codes[row_number]])
    # Backwards compatibility
    elif hasattr(df.index, 'labels'):
        for i, labels in enumerate(df.index.labels):
            result_index.append(df.index.levels[i][labels[row_number]])
    else:
        result_index.append(df.index[row_number])
    result_column = []
    if hasattr(df.columns, 'codes'):
        for i, codes in enumerate(df.columns.codes):
            result_column.append(df.columns.levels[i][codes[column_number]])
    # Backwards compatibility
    elif hasattr(df.columns, 'labels'):
        for i, labels in enumerate(df.columns.labels):
            result_column.append(df.columns.levels[i][labels[column_number]])
    else:
        result_column.append(df.columns[column_number])
    return result_index, result_column


def print_category_table(df):
    """
    Prints the category table to screen, from `Pandas DataFrame` input

    :param df: Pandas DataFrame input
    :type df: pandas.DataFrame
    """
    values = df.values  # data is converted to numpy array
    print("{:11s} {:10s} {:36s} {:20s}".format("Cell_ID", "Data", "Row Categories", "Column Categories"))
    for i, row in enumerate(values):
        for j, cell in enumerate(row):
            categories = find_multiindex_level(i, j, df)
            print("{:3} {:3} {:15}   {:35}  {:40}".format(i, j, str(cell), ''.join(str(categories[0])), ''.join(str(categories[1]))))


def build_category_table(df):
    """
    Builds the category table in form of a Python list, from `Pandas DataFrame` input

    :param df: Pandas DataFrame input
    :type df: pandas.DataFrame
    :return: category_table as Python list
    """
    values = df.values  # data is converted to numpy array
    category_table = []
    for i, row in enumerate(values):
        for j, cell in enumerate(row):
            data_point = []
            categories = find_multiindex_level(i, j, df)
            data_point.append(cell)
            data_point.append(categories[0])
            data_point.append(categories[1])
            category_table.append(data_point)
    return category_table



# -*- coding: utf-8 -*-
"""
Footnote handling.

.. codeauthor:: Juraj Mavračić <jm2111@cam.ac.uk>

"""

import logging
import numpy as np
import re
from .parse import CellParser

log = logging.getLogger(__name__)


class Footnote:
    """
    Defines a footnote found in the provided table.
    Contains elements of a footnote.
    Will construct the footnote and find all associated elements.

    :param table: table to work on
    :type table: ~tabledataextractor.table.table.Table
    :param prefix: Prefix that has been identified as footnote prefix
    :type prefix: str
    :param prefix_cell: Index of the cell containing the associated prefix
    :type prefix_cell: (int, int)
    :param text: Optional. Text associated with the found footnote prefix
    :type text: str
    """

    def __init__(self, table, prefix, prefix_cell, text):
        self._table = table
        self.pre_cleaned_table = np.copy(self._table.pre_cleaned_table)

        #: Prefix string, e.g., `"a)"`.
        self.prefix = prefix

        #: Cell index of the prefix, e.g., `(7,0)`.
        self.prefix_cell = prefix_cell

        #: Cell of the footnote text, e.g., `(7,1)`.
        self.text_cell = self.prefix_cell if text else self._find_text_cell()

        #: Footnote text, e.g., `"This is the text of a footnote"`.
        self.text = text if text else self._find_text()

        #: Cell indexes of the cells containing the footnote references within the table.
        self.reference_cells = self._find_reference_cells()

        #: Cell content of the cells contatining the footnote references within the table.
        self.references = self._find_references()

    def _find_text_cell(self):
        """Finds the cell index containing the text associated with the prefix."""
        for column_index in range(self.prefix_cell[1] + 1, np.shape(self.pre_cleaned_table)[1]):
            if not self._table.pre_cleaned_table_empty[self.prefix_cell[0], column_index]:
                return self.prefix_cell[0], column_index
            return None

    def _find_text(self):
        """Finds the text associated with the prefix, only one cell can contain the text."""
        if self.text_cell is not None:
            return str(self.pre_cleaned_table[self.text_cell])
        else:
            return ""

    def _find_reference_cells(self):
        """
        Searches the entire table above each footnote for the previously detected footnote prefix.
        Updates the footnote-internal version of the `pre-cleaned table`, by cutting out the footnote prefix out of the reference cell.

        Rules for matching:

            1. if prefix is `number`:
                a) matches if `(anything)+space+prefix`
            2. if prefix is `a-z`:
                a) matches if `(anything)+space+prefix` OR
                b) matches if `prefix`
            3. else:
                a) matches if found anywhere in any cell

        :return: [(int,int)]

        """
        # indices of the references
        fn_refs = []

        # Case 1a If prefix is number, general
        if re.fullmatch(pattern='[\d]{1,2}', string=self.prefix):
            log.debug("Footnote prefix {} is number".format(self.prefix))
            fn_ref_parser_1a = CellParser('(^.+\s)(' + self.prefix + ')(\s.+)?$')
            for fn_ref in fn_ref_parser_1a.parse(self.pre_cleaned_table[0:self.prefix_cell[0]], method='match'):
                fn_refs.append(fn_ref[:2])
                stripped_text = fn_ref[2][0]
                stripped_text += self.text if self.text is not None else ""
                if fn_ref[2][2] is not None:
                    stripped_text += fn_ref[2][2]
                self.pre_cleaned_table[fn_ref[:2]] = stripped_text

        # Case 2a If prefix is a-z:
        elif re.fullmatch(pattern='[a-zA-Z]', string=self.prefix):
            log.debug("Footnote prefix {} is letter".format(self.prefix))
            fn_ref_parser_2a = CellParser('(^.+\s)(' + self.prefix + ')(\s.+)?$')
            for fn_ref in fn_ref_parser_2a.parse(self.pre_cleaned_table[0:self.prefix_cell[0]], method='match'):
                fn_refs.append(fn_ref[:2])
                stripped_text = fn_ref[2][0]
                stripped_text += self.text if self.text is not None else ""
                if fn_ref[2][2] is not None:
                    stripped_text += fn_ref[2][2]
                self.pre_cleaned_table[fn_ref[:2]] = stripped_text

            # Case 2b If prefix is a-z and alone in the cell
            fn_ref_parser_2b = CellParser('^(' + self.prefix + ')$')
            for fn_ref in fn_ref_parser_2b.parse(self.pre_cleaned_table[0:self.prefix_cell[0]], method='match'):
                log.debug("Footnote prefix {} is letter and is alone in cell.".format(self.prefix))
                fn_refs.append(fn_ref[:2])
                stripped_text = self.text if self.text is not None else ""
                self.pre_cleaned_table[fn_ref[:2]] = stripped_text

        # Case 3, everything else
        else:
            fn_ref_parser = CellParser('(' + re.escape(self.prefix) + ')')
            repl = " "+self.text+" " if self.text is not None else " "
            for fn_ref in fn_ref_parser.replace(self.pre_cleaned_table[0:self.prefix_cell[0]],
                                                repl=repl,
                                                method='search'):
                fn_refs.append(fn_ref[:2])
                stripped_text = fn_ref[2]
                self.pre_cleaned_table[fn_ref[:2]] = stripped_text

        return fn_refs

    def _find_references(self):
        """Collects the raw references, no cleanup"""
        references = []
        for cell in self.reference_cells:
            references.append(self._table.pre_cleaned_table[cell])
        return references

    def __str__(self):
        return "Prefix: {:4}   Text: {:60}   Ref. Cells: {}   " \
               "References: {}".format("'"+str(self.prefix)+"'",
                                       "'"+str(self.text)+"'",
                                       str(self.reference_cells),
                                       str(self.references))


def find_footnotes(table_object):
    """
    Finds a footnote and yields a :class:`~tabledataextractor.table.footnotes.Footnote` object with all the appropriate properties.
    A footnote is defined with::

        FNprefix  = \*, #, ., o, †; possibly followed by "." or ")"

    A search is performed only below the data region.

    :param table_object: Input Table object
    :type table_object: ~tabledataextractor.table.table.Table
    """
    #: finds a footnote cell that possibly contains some text as well
    fn_parser = CellParser(r'^([*#\.o†\da-z][\.\)]?)(?!\d)\s?(([\w\[\]\s\:]+)?\.?)\s?$')
    for fn in fn_parser.parse(table_object.pre_cleaned_table):
        if fn[0] > table_object._cc4[0]:
            footnote = Footnote(table_object, prefix=fn[2][0], prefix_cell=(fn[0], fn[1]), text=fn[2][1])
            yield footnote




# -*- coding: utf-8 -*-
"""
Reads a `csv` formatted table from file. The file has to be 'utf-8' encoded.
"""

import numpy as np
import logging
import csv

log = logging.getLogger(__name__)


def read(file_path):
    """
    :param file_path: Path to `.csv` input file
    :type file_path: str
    :return: numpy.ndarray
    """

    with open(file_path, 'r', encoding='utf-8') as f:
        array = [elem for elem in list(csv.reader(f)) if elem]
        n = len(array[0])
        array = [x for x in array if len(x) == n]  # Only include rows with data for every column
        array = np.asarray(array, dtype='<U60')
    return array




# -*- coding: utf-8 -*-
"""
Functions for printing to screen in a nice format.
"""

import logging
import numpy as np
from prettytable import PrettyTable

log = logging.getLogger(__name__)


def print_table(table):
    """
    Prints a table to screen.

    :param table: input numpy array for printing
    :type table: numpy.array
    """
    n_columns = table.shape[1]
    cell_width = np.zeros(n_columns, dtype=int)

    # find the maximum cell width for each column i
    for i, column in enumerate(table.T):
        for cell in column:
            if len(cell) > cell_width[i]:
                cell_width[i] = len(cell)

    # print each row
    for row in table:
        for i, cell in enumerate(row):
            print("{:{cell_width}} ".format(cell, cell_width=cell_width[i]+1), end='', flush=True)
        print("\n", end='', flush=True)
    print("\n", flush=True)


def as_string(table):
    """
    Returns table as string for printing.

    :param table: input numpy array
    :type table: numpy.array
    :return: string
    """
    n_columns = table.shape[1]
    cell_width = np.zeros(n_columns, dtype=int)
    output = ''

    # find the maximum cell width for each column i
    for i, column in enumerate(table.T):
        for cell in column:
            if len(cell) > cell_width[i]:
                cell_width[i] = len(cell)

    # print each row
    for row in table:
        for i, cell in enumerate(row):
            output += "{:{cell_width}} ".format(cell, cell_width=cell_width[i]+1)
        output += "\n"
    output += "\n"
    return output


def list_as_PrettyTable(table_list):
    """
    Turns list into :class:`PrettyTable` object, ready for printing.

    :param table_list: list to be printed
    :return: table as `PrettyTable <https://pypi.org/project/PrettyTable/>`_

    """
    t = PrettyTable()
    t.field_names = ["Data", "Row Categories", "Column Categories"]
    for row in table_list:
        t.add_row(row)
    return t




# -*- coding: utf-8 -*-
"""
Represents a table in a highly standardized format.

.. codeauthor:: Juraj Mavračić <jm2111@cam.ac.uk>

"""

import logging
import numpy as np

from tabledataextractor.input import from_any
from tabledataextractor.output.print import as_string, print_table, list_as_PrettyTable
from tabledataextractor.output.to_csv import write_to_csv
from tabledataextractor.output.to_pandas import to_pandas, build_category_table
from tabledataextractor.table.parse import StringParser
from tabledataextractor.exceptions import InputError, MIPSError, TDEError
from tabledataextractor.table.history import History
from tabledataextractor.table.algorithms import find_cc1_cc2, find_cc3, find_cc4, prefix_duplicate_labels, \
    duplicate_spanning_cells, header_extension_up, find_title_row, find_note_cells, empty_cells, \
    pre_clean, split_table, standardize_empty, header_extension_down, find_row_header_table, clean_row_header
from tabledataextractor.table.footnotes import find_footnotes

log = logging.getLogger(__name__)


class Table:
    """
    Main `TableDataExtractor` object that includes the raw (input), cleaned (processes) and labelled tables.
    Represents the table input (.csv, .html, python list, url) in a highly standardized `category table` format,
    using the MIPS (*Minimum Indexing Point Search*) algorithm.

    Optional configuration keywords (defaults):

        * ``use_title_row = True``
            A title row will be assumed if possible.
        * ``use_prefixing = True``
            Will perform the prefixing steps if row or column index cells are not unique.
        * ``use_spanning_cells = True``
            Will duplicate spanning cells in the row and column header regions if needed.
        * ``use_header_extension = True``
            Will extend the row and column header beyond the MIPS-defined headers, if needed.
        * ``use_footnotes = True``
            Will copy the footnote text into the appropriate cells of the table and remove the footnote prefix.
        * ``use_max_data_area = False``
            If `True` the max data area will be used to determine the cell `CC2` in the main MIPS algorithm.
            It is probably never necessary to set this to True.
        * ``standardize_empty_data = True``
            Will standardize empty cells in the `data` region to 'NoValue'
        * ``row_header = None``
            If an integer is given, it indicates the index of `row_header` columns. This overwrites the MIPS algorithm.
            For example, ``row_header = 0`` will make only the first column a row header.
        * ``col_header = None``
            If an integer is given, it indicates the index of `col_header` rows. This overwrites the MIPS algorithm.
            For example, ``col_header = 0`` will make only the first row a column header.

    :param file_path: Path to .html or .cvs file, URL or list object that is used as input
    :type file_path: str | list
    :param table_number: Number of table to read, if there are several at the given url, or in the html file
    :type table_number: int
    """

    def __init__(self, file_path, table_number=1, **kwargs):
        """Runs required `TableDataExtractor` algorithms automatically upon initialization."""
        log.info('Initialization of table: "{}"'.format(file_path))
        self._file_path = file_path
        self._table_number = table_number
        self._configs = self._set_configs(**kwargs)
        self._history = History()
        self._analyze_table()

    @property
    def _default_configs(self):
        return {'use_title_row': True,
                'use_prefixing': True,
                'use_footnotes': True,
                'use_spanning_cells': True,
                'use_header_extension': True,
                'use_max_data_area': False,
                'standardize_empty_data': True,
                'row_header': None,
                'col_header': None}

    def _analyze_table(self):
        """
        Performs the analysis of the input table and is run automatically on initialization of the table object.
        """
        # check if input array is empty
        if empty_cells(self.raw_table).all():
            msg = 'Input table is empty.'
            log.critical(msg)
            raise InputError(msg)

        # clean-up the input array
        self._pre_cleaned_table = pre_clean(self.raw_table)
        log.debug("Table shape changed from {} to {}.".format(np.shape(self.raw_table),
                                                             np.shape(self.pre_cleaned_table)))

        if self.configs['use_spanning_cells']:
            self._pre_cleaned_table = duplicate_spanning_cells(self, self._pre_cleaned_table)

        if self.configs['use_prefixing']:
            self._pre_cleaned_table = prefix_duplicate_labels(self, self._pre_cleaned_table)

        # footnotes handling
        self._footnotes = []
        for footnote in find_footnotes(self):
            self._footnotes.append(footnote)
            if self.configs['use_footnotes']:
                self._copy_footnotes(footnote)

        # Main MIPS algorithm, finding the data and header regions
        try:
            #: Critical cells `CC1` and `CC2`
            self._cc1, self._cc2 = find_cc1_cc2(self, self._cc4, self._pre_cleaned_table)
        except (MIPSError, TypeError):
            msg = "ERROR: Main MIPS Algorithm failed. Maybe the input table is bad!"
            log.critical(msg)
            raise MIPSError(msg)
        else:
            log.debug("Table Cell CC1 = {}; Table Cell CC2 = {}".format(self._cc1, self._cc2))

        if self.configs['use_header_extension']:
            self._cc1 = header_extension_up(self, self._cc1)
            self._cc2 = header_extension_down(self, self._cc1, self._cc2, self._cc4)
            log.debug("Header extension, new cc1 = {}, new cc2 = {}".format(self._cc1, self._cc2))

        # check if critical cell `CC3` can be found
        try:
            _ = self._cc3
        except MIPSError:
            raise

    @property
    def footnotes(self):
        """
        List of footnotes in the table.
        Each footnote is an instance of :class:`~tabledataextractor.table.footnotes.Footnote`.

        :type: list[~tabledataextractor.table.footnotes.Footnote]
        """
        return self._footnotes

    @property
    def title_row(self):
        """
        Title row of the table.

        :type: list
        """
        if self._configs['use_title_row']:
            return find_title_row(self)

    @property
    def history(self):
        """
        Indicates which algorithms have been applied to the table by TableDataExtractor.

        :type: ~tabledataextractor.table.history.History
        """
        return self._history

    @property
    def labels(self):
        """
        Cell labels.

        :type: list
        """
        temp = np.empty_like(self._pre_cleaned_table, dtype="<U60")
        temp[:, :] = '/'

        if self.configs['use_title_row']:
            temp[self.title_row, :] = 'TableTitle'

        temp[self._cc1[0]:self._cc2[0] + 1, self._cc1[1]:self._cc2[1] + 1] = 'StubHeader'
        temp[self._cc3[0]:self._cc4[0] + 1, self._cc1[1]:self._cc2[1] + 1] = 'RowHeader'
        temp[self._cc1[0]:self._cc2[0] + 1, self._cc3[1]:self._cc4[1] + 1] = 'ColHeader'
        temp[self._cc3[0]:self._cc4[0] + 1, self._cc3[1]:self._cc4[1] + 1] = 'Data'

        for footnote in self.footnotes:
            temp[footnote.prefix_cell[0], footnote.prefix_cell[1]] = 'FNprefix'
            if footnote.text_cell is not None:
                temp[footnote.text_cell[0], footnote.text_cell[1]] = 'FNtext' if \
                    temp[footnote.text_cell[0], footnote.text_cell[1]] == '/' else 'FNprefix & FNtext'
            for ref_cell in footnote.reference_cells:
                temp[ref_cell[0], ref_cell[1]] = 'FNref' if temp[ref_cell[0], ref_cell[1]] == '/' else \
                    temp[ref_cell[0], ref_cell[1]] + ' & FNref'

        # all non-empty unlabelled cells at this point are labelled 'Note'
        for note_cell in find_note_cells(self, temp):
            temp[note_cell] = 'Note'
        return temp

    @property
    def configs(self):
        """
        Configuration keywords set at the creation of the :class:`~tabledataextractor.table.table.Table` instance.

        :type: dict
        """
        return self._configs

    @property
    def raw_table(self):
        """
        Input table, as provided to `TableDataExtractor`.

        :type: numpy.array
        """
        try:
            temp = from_any.create_table(self._file_path, self._table_number)
        except TypeError:
            raise
        else:
            assert isinstance(temp, np.ndarray) and temp.dtype == '<U60'
            if temp.ndim == 1:
                msg = 'Input table has only one row or column.'
                log.critical(msg)
                raise InputError(msg)
            if not self.history.table_transposed:
                return temp
            else:
                return temp.T

    @property
    def pre_cleaned_table(self):
        """
        Cleaned-up table.
        This table is used for labelling the table regions, finding data-cells and building the category table.

        :type: numpy.array
        """
        return self._pre_cleaned_table

    @property
    def pre_cleaned_table_empty(self):
        """
        Mask array with `True` for all empty cells of the ``pre_cleaned_table``.

        :type: numpy.array
        """
        return empty_cells(self._pre_cleaned_table)

    @property
    def category_table(self):
        """
        Standardized table, where each row corresponds to a single data point of the original table.
        The columns are the row and column categories where the data point belongs to.

        :type: list
        """
        if self._cc1 and self._cc2 and self._cc3 and self._cc4:
            return build_category_table(to_pandas(self))
        else:
            msg = "Category table not built. Critical cells have not been found."
            raise MIPSError(msg)

    @property
    def col_header(self):
        """
        Column header of the table.

        :type: numpy.ndarray
        """
        if self._cc1 and self._cc2 and self._cc3 and self._cc4:
            return self._pre_cleaned_table[self._cc1[0]:self._cc2[0] + 1, self._cc3[1]:self._cc4[1] + 1]
        else:
            msg = "No column header. Critical cells have not been found."
            raise MIPSError(msg)

    @property
    def row_header(self):
        """
        Row header of the table.

        :type: numpy.ndarray
        """
        if self._cc1 and self._cc2 and self._cc3 and self._cc4:
            return self._pre_cleaned_table[self._cc3[0]:self._cc4[0] + 1, self._cc1[1]:self._cc2[1] + 1]
        else:
            msg = "No row header. Critical cells have not been found."
            raise MIPSError(msg)

    @property
    def stub_header(self):
        """
        Stub header of the table.

        :type: numpy.ndarray
        """
        if self._cc1 and self._cc2 and self._cc3 and self._cc4:
            return self._pre_cleaned_table[self._cc1[0]:self._cc2[0] + 1, self._cc1[1]:self._cc2[1] + 1]
        else:
            msg = "No stub header. Critical cells have not been found."
            raise MIPSError(msg)

    @property
    def data(self):
        """
        Data region of the table.

        :type: numpy.ndarray
        """
        if self._cc1 and self._cc2 and self._cc3 and self._cc4:
            data_region = self._pre_cleaned_table[self._cc3[0]:self._cc4[0] + 1, self._cc3[1]:self._cc4[1] + 1]
            if self.configs['standardize_empty_data']:
                data_region = standardize_empty(data_region)
            return data_region
        else:
            msg = "No data region. Critical cells have not been found."
            raise MIPSError(msg)

    @property
    def subtables(self):
        """
        List of all subtables.
        Each subtable is an instance of :class:`~tabledataextractor.table.table.Table`.

        :type: list[~tabledataextractor.table.table.Table]
        """
        tables = []
        g = split_table(self)
        while True:
                subtable = next(g, None)
                if subtable is None:
                    break
                else:
                    try:
                        tables.append(Table(subtable))
                    except MIPSError as e:
                        log.exception("Subtable MIPS failure {}".format(e.args))
                        break
        return tables

    @property
    def row_categories(self):
        """
        Table where the original stub header is the first row(s) and all subsequent rows are the row categories of the
        original table. The assumption is made that the stub header labels row categories (that is, cells below the
        stub header). The `row_categories` table can be used if the row categories want to be analyzed as `data`
        themselves, which can occur if the header regions of the original table intentionally have duplicate elements.

        :type: ~tabledataextractor.table.table.TrivialTable
        """
        # this outer try statement is necessary to catch some weird errors with empty category tables
        try:
            if len(self.stub_header.T) != 0 and len(self.stub_header.T) == len(self.category_table[0][1]):
                raw_table = find_row_header_table(self.category_table, self.stub_header)
                try:
                    table = TrivialTable(raw_table, clean_row_header=True, row_header=0,
                                         col_header=len(self.stub_header) - 1)
                except TDEError as e:
                    return None
                if not empty_cells(table.data).any():
                    return table
            else:
                return None
        except IndexError:
            return None

    def contains(self, pattern):
        """
        Returns true if table contains a particular string.

        :param pattern: Regular expression for input
        :return: True/False
        """
        parser = StringParser(pattern)
        for row in self.category_table:
            string = row[0] + ' '
            string += ' '.join(row[1]) + ' '
            string += ' '.join(row[2])
            if parser.parse(string, method='search'):
                return True
        return False

    def transpose(self):
        """
        Transposes the `Table` and performs the analysis again.
        In this way, if working interactively from a `Jupyter` notebook, it is possible to input a table and then
        transpose it to see how it looks like and if the results of the standardization are different.
        """
        self._history = History()
        self.history._table_transposed = True
        self._analyze_table()

    @property
    def _cc4(self):
        """Critical cell `CC4`."""
        return find_cc4(self)

    @property
    def _cc3(self):
        """Critical cell `CC3`."""
        return find_cc3(self, self._cc2)

    def _set_configs(self, **kwargs):
        """Sets the configuration parameters based on the user input."""
        configs = self._default_configs
        for key, value in kwargs.items():
            if key in self._default_configs:
                configs[key] = value
            else:
                msg = 'Keyword "{}" does not exist.'.format(key)
                log.critical(msg)
                raise InputError(msg)
        log.info('Configuration parameters are: {}'.format(configs))
        return configs

    def _copy_footnotes(self, footnote):
        """
        Updates the pre-cleaned table with updated reference cells for a given footnote.
        """
        if not np.array_equal(self._pre_cleaned_table, footnote.pre_cleaned_table):
            self._pre_cleaned_table = np.copy(footnote.pre_cleaned_table)
            self.history._footnotes_copied = True
            log.debug("METHOD. Footnotes copied into cells.")

    def print(self):
        """
        Prints the `raw table` (input), `cleaned table` (processed by `TableDataExtractor`) and `labels`
        (regions of the table) nicely.
        """
        log.debug("Printing table: {}".format(self._file_path))
        print_table(self.raw_table)
        print_table(self._pre_cleaned_table)
        print_table(self.labels)

    def print_raw_table(self):
        """Prints raw input table nicely."""
        print_table(self.raw_table)

    def to_csv(self, file_path):
        """Saves the `raw_table` to a `.csv` file."""
        log.info("Saving raw table to .csv to file: {}".format(self._file_path))
        write_to_csv(self.raw_table, file_path=file_path)

    def to_pandas(self):
        """
        Converts the `Table` into a `Pandas DataFrame`, taking the complex MultiIndex structure of the table
        into account.

        :return: pandas.DataFrame
        """
        log.info("Converting table to Pandas DataFrame: {}".format(self._file_path))
        return to_pandas(self)

    def __str__(self):
        """As the user wants to see it"""
        log.debug("Printing table: {}".format(self._file_path))
        t = list_as_PrettyTable(self.category_table)
        return str(t)

    def __repr__(self):
        """As the developer wants to see it"""
        intro = "Table({}, table_number={}, transposed={})".format(self._file_path, self._table_number,
                                                                   self.history.table_transposed)
        log.debug("Repr. table: {}".format(self._file_path))
        array_width = np.shape(self._pre_cleaned_table)[1]
        input_string = as_string(self.raw_table)
        results_string = as_string(
            np.concatenate((self._pre_cleaned_table, np.full((1, array_width), "", dtype='<U60'), self.labels)))
        t = list_as_PrettyTable(self.category_table)
        return intro + "\n\n" + input_string + results_string + str(t)


class TrivialTable(Table):
    """
    Trivial Table object. No high level analysis will be performed. MIPS algorithm is never run.
    This table doesn't have footnotes, a title row or subtables.

    Optional configuration keywords (defaults):

        * ``standardize_empty_data = False``
            Will standardize empty cells in the `data` region to 'NoValue'.
        * ``clean_row_header = False``
            Removes duplicate rows that span the whole table (all columns).
        * ``row_header = 0``
            The column up to which the row header is defined.
        * ``col_header = 0``
            The row up to which the column header is defined.


    """
    def __init__(self, file_path, table_number=1, **kwargs):
        super().__init__(file_path=file_path, table_number=table_number, **kwargs)

    @property
    def _default_configs(self):
        return {'standardize_empty_data': False, 'clean_row_header': False, 'row_header': 0, 'col_header': 0}

    def _analyze_table(self):
        """
        Performs the analysis of the input table and is run automatically on initialization of the table object.
        """
        # check if input array is empty
        if empty_cells(self.raw_table).all():
            msg = 'Input table is empty.'
            log.critical(msg)
            raise InputError(msg)

        # define critical cells cc1 and cc2 (no MIPS algorithm is used in TrivialTable)
        self._cc1, self._cc2 = (0, 0), (self.configs['col_header'], self.configs['row_header'])
        log.debug("Table Cell CC1 = {}; Table Cell CC2 = {}".format(self._cc1, self._cc2))

        self._pre_cleaned_table = self.raw_table
        if self.configs['clean_row_header']:
            self._pre_cleaned_table = clean_row_header(self.pre_cleaned_table, self._cc2)

    @property
    def _cc4(self):
        """Critical cell `CC4`."""
        return len(self.pre_cleaned_table)-1, len(self.pre_cleaned_table.T)-1

    @property
    def _cc3(self):
        """Critical cell `CC3`."""
        if len(self.pre_cleaned_table.T) == 1:
            return self._cc2[0]+1, self._cc2[1]
        elif len(self.pre_cleaned_table) == 1:
            return self._cc2[0], self._cc2[1]+1
        else:
            return self._cc2[0]+1, self._cc2[1]+1

    @property
    def labels(self):
        """
        Cell labels.

        :type: numpy.array
        """
        temp = np.empty_like(self._pre_cleaned_table, dtype="<U60")
        temp[:, :] = '/'
        temp[self._cc1[0]:self._cc2[0] + 1, self._cc1[1]:self._cc2[1] + 1] = 'StubHeader'
        temp[self._cc3[0]:self._cc4[0] + 1, self._cc1[1]:self._cc2[1] + 1] = 'RowHeader'
        temp[self._cc1[0]:self._cc2[0] + 1, self._cc3[1]:self._cc4[1] + 1] = 'ColHeader'
        temp[self._cc3[0]:self._cc4[0] + 1, self._cc3[1]:self._cc4[1] + 1] = 'Data'
        return temp

    @property
    def col_header(self):
        """
        Column header of the table.

        :type: numpy.ndarray
        """
        if self._critical_cells and self._cc3[0] > self._cc2[0]:
            return self.pre_cleaned_table[self._cc1[0]:self._cc2[0] + 1, self._cc3[1]:self._cc4[1] + 1]
        elif self._critical_cells:
            return np.full_like(self.pre_cleaned_table[self._cc1[0]:self._cc2[0] + 1, self._cc3[1]:self._cc4[1] + 1],
                                fill_value='', dtype='<U60')
        else:
            return None

    @property
    def row_header(self):
        """
        Row header of the table. Enables a one-column table.

        :type: numpy.ndarray
        """
        if self._critical_cells and self._cc3[1] > self._cc2[1]:
            return self.pre_cleaned_table[self._cc3[0]:self._cc4[0] + 1, self._cc1[1]:self._cc2[1] + 1]
        elif self._critical_cells:
            return np.full_like(self.pre_cleaned_table[self._cc3[0]:self._cc4[0] + 1, self._cc1[1]:self._cc2[1] + 1],
                                fill_value='', dtype='<U60')
        else:
            return None

    @property
    def _critical_cells(self):
        """Indicates if all the critical cells have been found."""
        if self._cc1 and self._cc2 and self._cc3 and self._cc4:
            return True
        else:
            return False

    @property
    def footnotes(self):
        """None"""
        return None

    @property
    def title_row(self):
        """None"""
        return None

    @property
    def subtables(self):
        """None"""
        return None






# -*- coding: utf-8 -*-
"""
Exceptions defined for TableDataExtractor.
"""


class TDEError(Exception):
    """
    Base class for exceptions in TableDataExtractor.
    """


class InputError(TDEError):
    """
    Exception raised for errors in the input.
    """
    def __init__(self, message):
        self.message = message


class MIPSError(TDEError):
    """
    Exception raised for failure of the main MIPS algorithm.
    Usually signals that the table is broken or not well structured.
    """
    def __init__(self, message):
        self.message = message




from setuptools import setup, find_packages

setup(name='tabledataextractor',
      version='1.5.11',
      url='https://www.tabledataextractor.com',
      license='MIT License',
      author='Juraj Mavračić',
      author_email='jm2111@cam.ac.uk',
      description='Extracts data from tables',
      packages=find_packages(exclude=['tests']),
      long_description=open('README.md', 'r', encoding='utf-8').read(),
      long_description_content_type='text/markdown',
      zip_safe=False,
      test_suite='unittest',
      install_requires=[
            'django>2.1.6',
            'numpy==1.16; python_version < "3.7.0"',
            'numpy>=1.16,<2.0.0; python_version >= "3.7.0"',
            'sympy',
            'beautifulsoup4>=4.12.0',
            'requests>=2.21.0',
            'urllib3>=1.24.2',
            'selenium>=3.141.0',
            'prettytable>=0.7.2',
            'pandas==0.23.4; python_version < "3.7.0"',
            'pandas>=0.23.4; python_version >= "3.7.0"', ])



# -*- coding: utf-8 -*-
"""
Tests the spanning cells extension

.. codeauthor: Juraj Mavračić (jm2111@cam.ac.uk)
"""

import unittest
import logging

from tabledataextractor import Table

log = logging.getLogger(__name__)


class TestSpanningCells(unittest.TestCase):

    def test_default_config(self):
        table = Table("./tests/data/te_04.csv")
        table.print()
        pre_cleaned_table = [['Pupils in comprehensive schools and with leaving certificate', '', '', '', '', '', '', '', ''],
                             ['Year', 'School', 'Pupils', 'Pupils', 'Pupils', 'Pupils', 'Pupils', 'Grade 1','Leaving certificates'],
                             ['Year', 'School', 'Pre-primary', 'Grades', 'Grades', 'Additional', 'Total', 'Grade 1', 'Leaving certificates'],
                             ['Year', 'School', 'Pre-primary', '6 Jan', '9 Jul', 'Additional', 'Total', 'Grade 1', 'Leaving certificates'],
                             ['1990', '4869', '2189', '389410', '197719', '', '592920', '67427', '61054'],
                             ['1991', '4861', '2181', '389411', '197711', '3601', '592921', '67421', '']]
        category_table = [['4869', ['1990'], ['School', 'School', 'School']], ['2189', ['1990'], ['Pupils', 'Pre-primary', 'Pre-primary']], ['389410', ['1990'], ['Pupils', 'Grades', '6 Jan']], ['197719', ['1990'], ['Pupils', 'Grades', '9 Jul']], ['NoValue', ['1990'], ['Pupils', 'Additional', 'Additional']], ['592920', ['1990'], ['Pupils', 'Total', 'Total']], ['67427', ['1990'], ['Grade 1', 'Grade 1', 'Grade 1']], ['61054', ['1990'], ['Leaving certificates', 'Leaving certificates', 'Leaving certificates']], ['4861', ['1991'], ['School', 'School', 'School']], ['2181', ['1991'], ['Pupils', 'Pre-primary', 'Pre-primary']], ['389411', ['1991'], ['Pupils', 'Grades', '6 Jan']], ['197711', ['1991'], ['Pupils', 'Grades', '9 Jul']], ['3601', ['1991'], ['Pupils', 'Additional', 'Additional']], ['592921', ['1991'], ['Pupils', 'Total', 'Total']], ['67421', ['1991'], ['Grade 1', 'Grade 1', 'Grade 1']], ['NoValue', ['1991'], ['Leaving certificates', 'Leaving certificates', 'Leaving certificates']]]
        labels = [['TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle', 'TableTitle'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'], ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        self.assertListEqual(pre_cleaned_table, table._pre_cleaned_table.tolist())
        self.assertListEqual(category_table, table.category_table)
        self.assertListEqual(labels, table.labels.tolist())

    def test_specific_config(self):
        table = Table("./tests/data/te_04.csv", use_title_row=False, use_prefixing=False)
        table.print()
        pre_cleaned_table = [['Pupils in comprehensive schools and with leaving certificate', 'Pupils in comprehensive schools and with leaving certificate', 'Pupils in comprehensive schools and with leaving certificate', 'Pupils in comprehensive schools and with leaving certificate', 'Pupils in comprehensive schools and with leaving certificate', 'Pupils in comprehensive schools and with leaving certificate', 'Pupils in comprehensive schools and with leaving certificate', 'Pupils in comprehensive schools and with leaving certificate', 'Pupils in comprehensive schools and with leaving certificate'],
                             ['Year', 'School', 'Pupils', 'Pupils', 'Pupils', 'Pupils', 'Pupils', 'Grade 1','Leaving certificates'],
                             ['Year', 'School', 'Pre-primary', 'Grades', 'Grades', 'Additional', 'Total', 'Grade 1', 'Leaving certificates'],
                             ['Year', 'School', 'Pre-primary', '6 Jan', '9 Jul', 'Additional', 'Total', 'Grade 1', 'Leaving certificates'],
                             ['1990', '4869', '2189', '389410', '197719', '', '592920', '67427', '61054'],
                             ['1991', '4861', '2181', '389411', '197711', '3601', '592921', '67421', '']]
        category_table = [['4869', ['1990'], ['Pupils in comprehensive schools and with leaving certificate', 'School', 'School', 'School']], ['2189', ['1990'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Pre-primary', 'Pre-primary']], ['389410', ['1990'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Grades', '6 Jan']], ['197719', ['1990'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Grades', '9 Jul']], ['NoValue', ['1990'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Additional', 'Additional']], ['592920', ['1990'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Total', 'Total']], ['67427', ['1990'], ['Pupils in comprehensive schools and with leaving certificate', 'Grade 1', 'Grade 1', 'Grade 1']], ['61054', ['1990'], ['Pupils in comprehensive schools and with leaving certificate', 'Leaving certificates', 'Leaving certificates', 'Leaving certificates']], ['4861', ['1991'], ['Pupils in comprehensive schools and with leaving certificate', 'School', 'School', 'School']], ['2181', ['1991'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Pre-primary', 'Pre-primary']], ['389411', ['1991'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Grades', '6 Jan']], ['197711', ['1991'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Grades', '9 Jul']], ['3601', ['1991'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Additional', 'Additional']], ['592921', ['1991'], ['Pupils in comprehensive schools and with leaving certificate', 'Pupils', 'Total', 'Total']], ['67421', ['1991'], ['Pupils in comprehensive schools and with leaving certificate', 'Grade 1', 'Grade 1', 'Grade 1']], ['NoValue', ['1991'], ['Pupils in comprehensive schools and with leaving certificate', 'Leaving certificates', 'Leaving certificates', 'Leaving certificates']]]
        labels = [['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                  ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                  ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                  ['StubHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader', 'ColHeader'],
                  ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data'],
                  ['RowHeader', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data', 'Data']]
        self.assertListEqual(pre_cleaned_table, table._pre_cleaned_table.tolist())
        self.assertListEqual(category_table, table.category_table)
        self.assertListEqual(labels, table.labels.tolist())


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
"""
tabledataextractor.input

Methods for input and interfacing with html,xml,cls,pdf readers.
The output of these methods is always a Numpy array.

jm2111@cam.ac.uk
~~~~~~~~~~~~~~~~~
"""










# -*- coding: utf-8 -*-
"""
Outputs the table to cvs.
"""

import logging
import csv
import os

log = logging.getLogger(__name__)


def write_to_csv(table, file_path):
    """
    Writes a numpy array table to a .csv file.
    Overrides existing files.

    :param table: Array of table data
    :type table: ndarray
    :param file_path: Output location
    :type file_path: str
    """
    if os.path.exists(file_path):
        log.info("File: {} overwritten.".format(file_path))
    with open(file_path, 'w', encoding='utf-8') as f:
        csv.writer(f).writerows(table)
